@@ -1,26 +1,67 @@
 #!/usr/bin/env python
-# -*- encoding: utf-8 -*-
 import logging
-import re
+import os
+import sys
+from types import TracebackType
+from typing import Any
+from typing import cast
+from typing import Dict
+from typing import List
+from typing import Optional
+from typing import Tuple
+from typing import TYPE_CHECKING
+from typing import Union
 from urllib.parse import unquote
 
 import requests
-from caldav.elements import cdav
+from lxml import etree
+from lxml.etree import _Element
+from requests.auth import AuthBase
+from requests.models import Response
+from requests.structures import CaseInsensitiveDict
+
+from .elements.base import BaseElement
+from caldav import __version__
 from caldav.elements import dav
-from caldav.elements import ical
 from caldav.lib import error
 from caldav.lib.python_utilities import to_normal_str
-from caldav.lib.python_utilities import to_unicode
 from caldav.lib.python_utilities import to_wire
 from caldav.lib.url import URL
 from caldav.objects import Calendar
-from caldav.objects import errmsg
 from caldav.objects import log
 from caldav.objects import Principal
-from caldav.objects import ScheduleInbox
-from caldav.objects import ScheduleOutbox
 from caldav.requests import HTTPBearerAuth
-from lxml import etree
+
+if TYPE_CHECKING:
+    pass
+
+if sys.version_info < (3, 9):
+    from typing import Iterable, Mapping
+else:
+    from collections.abc import Iterable, Mapping
+
+if sys.version_info < (3, 11):
+    from typing_extensions import Self
+else:
+    from typing import Self
+
+## TODO: this is also declared in davclient.DAVClient.__init__(...)
+## TODO: it should be consolidated, duplication is a bad thing
+## TODO: and it's almost certain that we'll forget to update this list
+CONNKEYS = set(
+    (
+        "url",
+        "proxy",
+        "username",
+        "password",
+        "timeout",
+        "headers",
+        "huge_tree",
+        "ssl_verify_cert",
+        "ssl_cert",
+        "auth",
+    )
+)
 
 
 class DAVResponse:
@@ -32,78 +73,89 @@ class DAVResponse:
     """
 
     raw = ""
-    reason = ""
-    tree = None
-    headers = {}
-    status = 0
+    reason: str = ""
+    tree: Optional[_Element] = None
+    headers: CaseInsensitiveDict = None
+    status: int = 0
+    davclient = None
+    huge_tree: bool = False
 
-    def __init__(self, response):
+    def __init__(
+        self, response: Response, davclient: Optional["DAVClient"] = None
+    ) -> None:
         self.headers = response.headers
         log.debug("response headers: " + str(self.headers))
         log.debug("response status: " + str(self.status))
 
         self._raw = response.content
+        self.davclient = davclient
+        if davclient:
+            self.huge_tree = davclient.huge_tree
 
-        ## TODO: this if/else/elif could possibly be refactored, or we should
-        ## consider to do streaming into the xmltree library as originally
-        ## intended.  It only makes sense for really huge payloads though.
-        if self.headers.get("Content-Type", "").startswith(
-            "text/xml"
-        ) or self.headers.get("Content-Type", "").startswith("application/xml"):
-            try:
-                content_length = int(self.headers["Content-Length"])
-            except:
-                content_length = -1
-            if content_length == 0 or not self._raw:
-                self._raw = ""
-                self.tree = None
-                log.debug("No content delivered")
-            else:
-                ## With response.raw we could be streaming the content, but it does not work because
-                ## the stream often is compressed.  We could add uncompression on the fly, but not
-                ## considered worth the effort as for now.
-                # self.tree = etree.parse(response.raw, parser=etree.XMLParser(remove_blank_text=True))
-                try:
-                    self.tree = etree.XML(
-                        self._raw, parser=etree.XMLParser(remove_blank_text=True)
-                    )
-                except:
-                    logging.critical(
-                        "Expected some valid XML from the server, but got this: \n"
-                        + self._raw,
-                        exc_info=True,
-                    )
-                    raise
-                if log.level <= logging.DEBUG:
-                    log.debug(etree.tostring(self.tree, pretty_print=True))
-        elif self.headers.get("Content-Type", "").startswith(
-            "text/calendar"
-        ) or self.headers.get("Content-Type", "").startswith("text/plain"):
-            ## text/plain is typically for errors, we shouldn't see it on 200/207 responses.
-            ## TODO: may want to log an error if it's text/plain and 200/207.
-            ## Logic here was moved when refactoring
-            pass
+        content_type = self.headers.get("Content-Type", "")
+        xml = ["text/xml", "application/xml"]
+        no_xml = ["text/plain", "text/calendar", "application/octet-stream"]
+        expect_xml = any((content_type.startswith(x) for x in xml))
+        expect_no_xml = any((content_type.startswith(x) for x in no_xml))
+        if content_type and not expect_xml and not expect_no_xml:
+            error.weirdness(f"Unexpected content type: {content_type}")
+        try:
+            content_length = int(self.headers["Content-Length"])
+        except:
+            content_length = -1
+        if content_length == 0 or not self._raw:
+            self._raw = ""
+            self.tree = None
+            log.debug("No content delivered")
         else:
-            ## Probably no content type given (iCloud).  Some servers
-            ## give text/html as the default when no content is
-            ## delivered or on errors (ref
-            ## https://github.com/python-caldav/caldav/issues/142).
-            ## TODO: maybe just remove all of the code above in this if/else and let all
-            ## data be parsed through this code.
+            ## For really huge objects we should pass the object as a stream to the
+            ## XML parser, like this:
+            # self.tree = etree.parse(response.raw, parser=etree.XMLParser(remove_blank_text=True))
+            ## However, we would also need to decompress on the fly.  I won't bother now.
             try:
+                ## https://github.com/python-caldav/caldav/issues/142
+                ## We cannot trust the content=type (iCloud, OX and others).
+                ## We'll try to parse the content as XML no matter
+                ## the content type given.
                 self.tree = etree.XML(
-                    self._raw, parser=etree.XMLParser(remove_blank_text=True)
+                    self._raw,
+                    parser=etree.XMLParser(
+                        remove_blank_text=True, huge_tree=self.huge_tree
+                    ),
                 )
             except:
-                pass
+                ## Content wasn't XML.  What does the content-type say?
+                ## expect_no_xml means text/plain or text/calendar
+                ## expect_no_xml -> ok, pass on, with debug logging
+                ## expect_xml means text/xml or application/xml
+                ## expect_xml -> raise an error
+                ## anything else (text/plain, text/html, ''),
+                ## log an error and continue
+                if not expect_no_xml or log.level <= logging.DEBUG:
+                    if not expect_no_xml:
+                        _log = logging.critical
+                    else:
+                        _log = logging.debug
+                        ## The statement below may not be true.
+                        ## We may be expecting something else
+                    _log(
+                        "Expected some valid XML from the server, but got this: \n"
+                        + str(self._raw),
+                        exc_info=True,
+                    )
+                if expect_xml:
+                    raise
+            else:
+                if log.level <= logging.DEBUG:
+                    log.debug(etree.tostring(self.tree, pretty_print=True))
 
         ## this if will always be true as for now, see other comments on streaming.
         if hasattr(self, "_raw"):
             log.debug(self._raw)
             # ref https://github.com/python-caldav/caldav/issues/112 stray CRs may cause problems
-            if type(self._raw) == bytes:
+            if isinstance(self._raw, bytes):
                 self._raw = self._raw.replace(b"\r\n", b"\n")
-            elif type(self._raw) == str:
+            elif isinstance(self._raw, str):
                 self._raw = self._raw.replace("\r\n", "\n")
         self.status = response.status_code
         ## ref https://github.com/python-caldav/caldav/issues/81,
@@ -115,11 +167,11 @@ class DAVResponse:
             self.reason = ""
 
     @property
-    def raw(self):
+    def raw(self) -> str:
         ## TODO: this should not really be needed?
         if not hasattr(self, "_raw"):
-            self._raw = etree.tostring(self.tree, pretty_print=True)
-        return self._raw
+            self._raw = etree.tostring(cast(_Element, self.tree), pretty_print=True)
+        return to_normal_str(self._raw)
 
     def _strip_to_multistatus(self):
         """
@@ -150,7 +202,7 @@ class DAVResponse:
             return self.tree
         return [self.tree]
 
-    def validate_status(self, status):
+    def validate_status(self, status: str) -> None:
         """
         status is a string like "HTTP/1.1 404 Not Found".  200, 207 and
         404 are considered good statuses.  The SOGo caldav server even
@@ -167,15 +219,16 @@ class DAVResponse:
         ):
             raise error.ResponseError(status)
 
-    def _parse_response(self, response):
+    def _parse_response(self, response) -> Tuple[str, List[_Element], Optional[Any]]:
         """
         One response should contain one or zero status children, one
         href tag and zero or more propstats.  Find them, assert there
         isn't more in the response and return those three fields
         """
         status = None
-        href = None
-        propstats = []
+        href: Optional[str] = None
+        propstats: List[_Element] = []
+        check_404 = False  ## special for purelymail
         error.assert_(response.tag == dav.Response.tag)
         for elem in response:
             if elem.tag == dav.Status.tag:
@@ -185,15 +238,41 @@ class DAVResponse:
                 self.validate_status(status)
             elif elem.tag == dav.Href.tag:
                 assert not href
+                # Fix for https://github.com/python-caldav/caldav/issues/471
+                # Confluence server quotes the user email twice. We unquote it manually.
+                if "%2540" in elem.text:
+                    elem.text = elem.text.replace("%2540", "%40")
                 href = unquote(elem.text)
             elif elem.tag == dav.PropStat.tag:
                 propstats.append(elem)
+            elif elem.tag == "{DAV:}error":
+                ## This happens with purelymail on a 404.
+                ## This code is mostly moot, but in debug
+                ## mode I want to be sure we do not toss away any data
+                children = elem.getchildren()
+                error.assert_(len(children) == 1)
+                error.assert_(
+                    children[0].tag == "{https://purelymail.com}does-not-exist"
+                )
+                check_404 = True
             else:
-                error.assert_(False)
+                ## i.e. purelymail may contain one more tag, <error>...</error>
+                ## This is probably not a breach of the standard.  It may
+                ## probably be ignored.  But it's something we may want to
+                ## know.
+                error.weirdness("unexpected element found in response", elem)
         error.assert_(href)
-        return (href, propstats, status)
+        if check_404:
+            error.assert_("404" in status)
+        ## TODO: is this safe/sane?
+        ## Ref https://github.com/python-caldav/caldav/issues/435 the paths returned may be absolute URLs,
+        ## but the caller expects them to be paths.  Could we have issues when a server has same path
+        ## but different URLs for different elements?  Perhaps href should always be made into an URL-object?
+        if ":" in href:
+            href = unquote(URL(href).path)
+        return (cast(str, href), propstats, status)
 
-    def find_objects_and_props(self):
+    def find_objects_and_props(self) -> Dict[str, Dict[str, _Element]]:
         """Check the response from the server, check that it is on an expected format,
         find hrefs and props from it and check statuses delivered.
 
@@ -203,7 +282,8 @@ class DAVResponse:
 
         self.sync_token will be populated if found, self.objects will be populated.
         """
-        self.objects = {}
+        self.objects: Dict[str, Dict[str, _Element]] = {}
+        self.statuses: Dict[str, str] = {}
 
         if "Schedule-Tag" in self.headers:
             self.schedule_tag = self.headers["Schedule-Tag"]
@@ -219,8 +299,9 @@ class DAVResponse:
             ## I would like to do this assert here ...
             # error.assert_(not href in self.objects)
             ## but then there was https://github.com/python-caldav/caldav/issues/136
-            if not href in self.objects:
+            if href not in self.objects:
                 self.objects[href] = {}
+                self.statuses[href] = status
 
             ## The properties may be delivered either in one
             ## propstat with multiple props or in multiple
@@ -229,7 +310,7 @@ class DAVResponse:
                 cnt = 0
                 status = propstat.find(dav.Status.tag)
                 error.assert_(status is not None)
-                if status is not None:
+                if status is not None and status.text is not None:
                     error.assert_(len(status) == 0)
                     cnt += 1
                     self.validate_status(status.text)
@@ -252,7 +333,12 @@ class DAVResponse:
         values = []
         if proptag in props_found:
             prop_xml = props_found[proptag]
-            error.assert_(not prop_xml.items())
+            if prop_xml.items():
+                from caldav.lib.debug import xmlstring
+
+                log.error(
+                    f"If you see this, please add a report at https://github.com/python-caldav/caldav/issues/209 - in _expand_simple_prop, dealing with {proptag}, extra items found: {xmlstring(prop_xml)}."
+                )
             if not xpath and len(prop_xml) == 0:
                 if prop_xml.text:
                     values.append(prop_xml.text)
@@ -274,8 +360,13 @@ class DAVResponse:
             error.assert_(len(values) == 1)
             return values[0]
 
-    ## TODO: "expand" does not feel quite right.
-    def expand_simple_props(self, props=[], multi_value_props=[], xpath=None):
+    ## TODO: word "expand" does not feel quite right.
+    def expand_simple_props(
+        self,
+        props: Iterable[BaseElement] = None,
+        multi_value_props: Iterable[Any] = None,
+        xpath: Optional[str] = None,
+    ) -> Dict[str, Dict[str, str]]:
         """
         The find_objects_and_props() will stop at the xml element
         below the prop tag.  This method will expand those props into
@@ -284,19 +375,29 @@ class DAVResponse:
         Executes find_objects_and_props if not run already, then
         modifies and returns self.objects.
         """
+        props = props or []
+        multi_value_props = multi_value_props or []
+
         if not hasattr(self, "objects"):
             self.find_objects_and_props()
         for href in self.objects:
             props_found = self.objects[href]
             for prop in props:
+                if prop.tag is None:
+                    continue
+
                 props_found[prop.tag] = self._expand_simple_prop(
                     prop.tag, props_found, xpath=xpath
                 )
             for prop in multi_value_props:
+                if prop.tag is None:
+                    continue
+
                 props_found[prop.tag] = self._expand_simple_prop(
                     prop.tag, props_found, xpath=xpath, multi_value_allowed=True
                 )
-        return self.objects
+        # _Element objects in self.objects are parsed to str, thus the need to cast the return
+        return cast(Dict[str, Dict[str, str]], self.objects)
 
 
 class DAVClient:
@@ -308,21 +409,23 @@ class DAVClient:
     the constructor (__init__), the principal method and the calendar method.
     """
 
-    proxy = None
-    url = None
+    proxy: Optional[str] = None
+    url: URL = None
+    huge_tree: bool = False
 
     def __init__(
         self,
-        url,
-        proxy=None,
-        username=None,
-        password=None,
-        auth=None,
-        timeout=None,
-        ssl_verify_cert=True,
-        ssl_cert=None,
-        headers={},
-    ):
+        url: str,
+        proxy: Optional[str] = None,
+        username: Optional[str] = None,
+        password: Optional[str] = None,
+        auth: Optional[AuthBase] = None,
+        timeout: Optional[int] = None,
+        ssl_verify_cert: Union[bool, str] = True,
+        ssl_cert: Union[str, Tuple[str, str], None] = None,
+        headers: Mapping[str, str] = None,
+        huge_tree: bool = False,
+    ) -> None:
         """
         Sets up a HTTPConnection object towards the server in the url.
         Parameters:
@@ -330,42 +433,47 @@ class DAVClient:
          * proxy: A string defining a proxy server: `hostname:port`
          * username and password should be passed as arguments or in the URL
          * auth, timeout and ssl_verify_cert are passed to requests.request.
-         ** ssl_verify_cert can be the path of a CA-bundle or False.
+         * ssl_verify_cert can be the path of a CA-bundle or False.
+         * huge_tree: boolean, enable XMLParser huge_tree to handle big events, beware
+           of security issues, see : https://lxml.de/api/lxml.etree.XMLParser-class.html
 
         The requests library will honor a .netrc-file, if such a file exists
         username and password may be omitted.  Known bug: .netrc is honored
         even if a username/password is given, ref https://github.com/python-caldav/caldav/issues/206
         """
+        headers = headers or {}
 
         self.session = requests.Session()
 
         log.debug("url: " + str(url))
         self.url = URL.objectify(url)
-
+        self.huge_tree = huge_tree
         # Prepare proxy info
         if proxy is not None:
-            self.proxy = proxy
+            _proxy = proxy
             # requests library expects the proxy url to have a scheme
-            if re.match("^.*://", proxy) is None:
-                self.proxy = self.url.scheme + "://" + proxy
+            if "://" not in proxy:
+                _proxy = self.url.scheme + "://" + proxy
 
             # add a port is one is not specified
             # TODO: this will break if using basic auth and embedding
             # username:password in the proxy URL
-            p = self.proxy.split(":")
+            p = _proxy.split(":")
             if len(p) == 2:
-                self.proxy += ":8080"
-            log.debug("init - proxy: %s" % (self.proxy))
+                _proxy += ":8080"
+            log.debug("init - proxy: %s" % (_proxy))
+
+            self.proxy = _proxy
 
         # Build global headers
-        self.headers = headers
-        self.headers.update(
+        self.headers = CaseInsensitiveDict(
             {
-                "User-Agent": "Mozilla/5.0",
+                "User-Agent": "python-caldav/" + __version__,
                 "Content-Type": "text/xml",
                 "Accept": "text/xml, text/calendar",
             }
         )
+        self.headers.update(headers or {})
         if self.url.username is not None:
             username = unquote(self.url.username)
             password = unquote(self.url.password)
@@ -373,7 +481,7 @@ class DAVClient:
         self.username = username
         self.password = password
         ## I had problems with passwords with non-ascii letters in it ...
-        if hasattr(self.password, "encode"):
+        if isinstance(self.password, str):
             self.password = self.password.encode("utf-8")
         self.auth = auth
         # TODO: it's possible to force through a specific auth method here,
@@ -386,13 +494,18 @@ class DAVClient:
 
         self._principal = None
 
-    def __enter__(self):
+    def __enter__(self) -> Self:
         return self
 
-    def __exit__(self, exc_type, exc_value, traceback):
+    def __exit__(
+        self,
+        exc_type: Optional[BaseException],
+        exc_value: Optional[BaseException],
+        traceback: Optional[TracebackType],
+    ) -> None:
         self.close()
 
-    def close(self):
+    def close(self) -> None:
         """
         Closes the DAVClient's session object
         """
@@ -414,7 +527,7 @@ class DAVClient:
     def calendar(self, **kwargs):
         """Returns a calendar object.
 
-        Typically, an URL should be given as a named parameter (url)
+        Typically, a URL should be given as a named parameter (url)
 
         No network traffic will be initiated by this method.
 
@@ -424,7 +537,7 @@ class DAVClient:
         """
         return Calendar(client=self, **kwargs)
 
-    def check_dav_support(self):
+    def check_dav_support(self) -> Optional[str]:
         try:
             ## SOGo does not return the full capability list on the caldav
             ## root URL, and that's OK according to the RFC ... so apparently
@@ -433,18 +546,20 @@ class DAVClient:
             ## Anyway, packing this into a try-except in case it fails.
             response = self.options(self.principal().url)
         except:
-            response = self.options(self.url)
+            response = self.options(str(self.url))
         return response.headers.get("DAV", None)
 
-    def check_cdav_support(self):
+    def check_cdav_support(self) -> bool:
         support_list = self.check_dav_support()
-        return support_list and "calendar-access" in support_list
+        return support_list is not None and "calendar-access" in support_list
 
-    def check_scheduling_support(self):
+    def check_scheduling_support(self) -> bool:
         support_list = self.check_dav_support()
-        return support_list and "calendar-auto-schedule" in support_list
+        return support_list is not None and "calendar-auto-schedule" in support_list
 
-    def propfind(self, url=None, props="", depth=0):
+    def propfind(
+        self, url: Optional[str] = None, props: str = "", depth: int = 0
+    ) -> DAVResponse:
         """
         Send a propfind request.
 
@@ -456,9 +571,11 @@ class DAVClient:
         Returns
          * DAVResponse
         """
-        return self.request(url or self.url, "PROPFIND", props, {"Depth": str(depth)})
+        return self.request(
+            url or str(self.url), "PROPFIND", props, {"Depth": str(depth)}
+        )
 
-    def proppatch(self, url, body, dummy=None):
+    def proppatch(self, url: str, body: str, dummy: None = None) -> DAVResponse:
         """
         Send a proppatch request.
 
@@ -472,7 +589,7 @@ class DAVClient:
         """
         return self.request(url, "PROPPATCH", body)
 
-    def report(self, url, query="", depth=0):
+    def report(self, url: str, query: str = "", depth: int = 0) -> DAVResponse:
         """
         Send a report request.
 
@@ -491,7 +608,7 @@ class DAVClient:
             {"Depth": str(depth), "Content-Type": 'application/xml; charset="utf-8"'},
         )
 
-    def mkcol(self, url, body, dummy=None):
+    def mkcol(self, url: str, body: str, dummy: None = None) -> DAVResponse:
         """
         Send a MKCOL request.
 
@@ -514,7 +631,7 @@ class DAVClient:
         """
         return self.request(url, "MKCOL", body)
 
-    def mkcalendar(self, url, body="", dummy=None):
+    def mkcalendar(self, url: str, body: str = "", dummy: None = None) -> DAVResponse:
         """
         Send a mkcalendar request.
 
@@ -528,60 +645,70 @@ class DAVClient:
         """
         return self.request(url, "MKCALENDAR", body)
 
-    def put(self, url, body, headers={}):
+    def put(
+        self, url: str, body: str, headers: Mapping[str, str] = None
+    ) -> DAVResponse:
         """
         Send a put request.
         """
-        return self.request(url, "PUT", body, headers)
+        return self.request(url, "PUT", body, headers or {})
 
-    def post(self, url, body, headers={}):
+    def post(
+        self, url: str, body: str, headers: Mapping[str, str] = None
+    ) -> DAVResponse:
         """
         Send a POST request.
         """
-        return self.request(url, "POST", body, headers)
+        return self.request(url, "POST", body, headers or {})
 
-    def delete(self, url):
+    def delete(self, url: str) -> DAVResponse:
         """
         Send a delete request.
         """
         return self.request(url, "DELETE")
 
-    def options(self, url):
+    def options(self, url: str) -> DAVResponse:
         return self.request(url, "OPTIONS")
 
-    def extract_auth_types(self, header):
-        auth_types = header.lower().split(",")
-        auth_types = map(lambda auth_type: auth_type.strip(), auth_types)
-        auth_types = map(lambda auth_type: auth_type.split(" ")[0], auth_types)
-        return list(filter(lambda auth_type: auth_type, auth_types))
+    def extract_auth_types(self, header: str):
+        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/WWW-Authenticate#syntax
+        return {h.split()[0] for h in header.lower().split(",")}
 
-    def request(self, url, method="GET", body="", headers={}):
+    def request(
+        self,
+        url: str,
+        method: str = "GET",
+        body: str = "",
+        headers: Mapping[str, str] = None,
+    ) -> DAVResponse:
         """
-        Actually sends the request
+        Actually sends the request, and does the authentication
         """
+        headers = headers or {}
+
         combined_headers = self.headers.copy()
-        combined_headers.update(headers)
+        combined_headers.update(headers or {})
         if (body is None or body == "") and "Content-Type" in combined_headers:
             del combined_headers["Content-Type"]
 
+        # objectify the url
+        url_obj = URL.objectify(url)
+
         proxies = None
         if self.proxy is not None:
-            proxies = {url.scheme: self.proxy}
+            proxies = {url_obj.scheme: self.proxy}
             log.debug("using proxy - %s" % (proxies))
 
-        # objectify the url
-        url = URL.objectify(url)
-
         log.debug(
             "sending request - method={0}, url={1}, headers={2}\nbody:\n{3}".format(
-                method, str(url), combined_headers, to_normal_str(body)
+                method, str(url_obj), combined_headers, to_normal_str(body)
             )
         )
 
         try:
             r = self.session.request(
                 method,
-                str(url),
+                str(url_obj),
                 data=to_wire(body),
                 headers=combined_headers,
                 proxies=proxies,
@@ -591,7 +718,7 @@ class DAVClient:
                 cert=self.ssl_cert,
             )
             log.debug("server responded with %i %s" % (r.status_code, r.reason))
-            response = DAVResponse(r)
+            response = DAVResponse(r, self)
         except:
             ## this is a workaround needed due to some weird server
             ## that would just abort the connection rather than send a
@@ -601,7 +728,7 @@ class DAVClient:
                 raise
             r = self.session.request(
                 method="GET",
-                url=str(url),
+                url=str(url_obj),
                 headers=combined_headers,
                 proxies=proxies,
                 timeout=self.timeout,
@@ -611,47 +738,65 @@ class DAVClient:
             if not r.status_code == 401:
                 raise
 
-        if r.status_code == 401 and "WWW-Authenticate" in r.headers and not self.auth:
-            auth_types = self.extract_auth_types(r.headers["WWW-Authenticate"])
-
-            if self.password and self.username and "digest" in auth_types:
+        ## Returned headers
+        r_headers = CaseInsensitiveDict(r.headers)
+        if (
+            r.status_code == 401
+            and "WWW-Authenticate" in r_headers
+            and not self.auth
+            and self.username
+        ):
+            auth_types = self.extract_auth_types(r_headers["WWW-Authenticate"])
+            if self.username and "digest" in auth_types:
                 self.auth = requests.auth.HTTPDigestAuth(self.username, self.password)
-            elif self.password and self.username and "basic" in auth_types:
+            elif self.username and "basic" in auth_types:
                 self.auth = requests.auth.HTTPBasicAuth(self.username, self.password)
             elif self.password and "bearer" in auth_types:
                 self.auth = HTTPBearerAuth(self.password)
+            elif "bearer" in auth_types:
+                raise error.AuthorizationError(
+                    reason="Server provides bearer auth, but no password given.  The bearer token should be configured as password"
+                )
             else:
                 raise NotImplementedError(
                     "The server does not provide any of the currently "
-                    "supported authentication methods: basic, digest"
+                    "supported authentication methods: basic, digest, bearer"
                 )
 
             return self.request(url, method, body, headers)
 
-        elif r.status_code == 401 and "WWW-Authenticate" in r.headers and self.auth:
+        elif (
+            r.status_code == 401
+            and "WWW-Authenticate" in r_headers
+            and self.auth
+            and self.password
+            and isinstance(self.password, bytes)
+        ):
+            ## Most likely we're here due to wrong username/password
+            ## combo, but it could also be charset problems.  Some
+            ## (ancient) servers don't like UTF-8 binary auth with
+            ## Digest authentication.  An example are old SabreDAV
+            ## based servers.  Not sure about UTF-8 and Basic Auth,
+            ## but likely the same.  so retry if password is a bytes
+            ## sequence and not a string (see commit 13a4714, which
+            ## introduced this regression)
 
-            ## Some (ancient) servers don't like UTF-8 binary auth with Digest authentication.
-            ## An example are old SabreDAV based servers.  Not sure about UTF-8 and Basic Auth,
-            ## but likely the same.  so retry if password is a bytes sequence and not a string
-            ## (see commit 13a4714, which introduced this regression)
+            auth_types = self.extract_auth_types(r_headers["WWW-Authenticate"])
 
-            auth_types = self.extract_auth_types(r.headers["WWW-Authenticate"])
-
-            if self.password and hasattr(self.password, "decode"):
-                if self.username and "digest" in auth_types:
-                    self.auth = requests.auth.HTTPDigestAuth(
-                        self.username, self.password.decode()
-                    )
-                elif self.username and "basic" in auth_types:
-                    self.auth = requests.auth.HTTPBasicAuth(
-                        self.username, self.password.decode()
-                    )
-                elif "bearer" in auth_types:
-                    self.auth = HTTPBearerAuth(self.password.decode())
+            if self.password and self.username and "digest" in auth_types:
+                self.auth = requests.auth.HTTPDigestAuth(
+                    self.username, self.password.decode()
+                )
+            elif self.password and self.username and "basic" in auth_types:
+                self.auth = requests.auth.HTTPBasicAuth(
+                    self.username, self.password.decode()
+                )
+            elif self.password and "bearer" in auth_types:
+                self.auth = HTTPBearerAuth(self.password.decode())
 
             self.username = None
             self.password = None
-            return self.request(url, method, body, headers)
+            return self.request(str(url_obj), method, body, headers)
 
         # this is an error condition that should be raised to the application
         if (
@@ -662,6 +807,117 @@ class DAVClient:
                 reason = response.reason
             except AttributeError:
                 reason = "None given"
-            raise error.AuthorizationError(url=str(url), reason=reason)
+            raise error.AuthorizationError(url=str(url_obj), reason=reason)
+
+        if error.debug_dump_communication:
+            import datetime
+            from tempfile import NamedTemporaryFile
+
+            with NamedTemporaryFile(prefix="caldavcomm", delete=False) as commlog:
+                commlog.write(b"=" * 80 + b"\n")
+                commlog.write(f"{datetime.datetime.now():%FT%H:%M:%S}".encode("utf-8"))
+                commlog.write(b"\n====>\n")
+                commlog.write(f"{method} {url}\n".encode("utf-8"))
+                commlog.write(
+                    b"\n".join(to_wire(f"{x}: {headers[x]}") for x in headers)
+                )
+                commlog.write(b"\n\n")
+                commlog.write(to_wire(body))
+                commlog.write(b"<====\n")
+                commlog.write(f"{response.status} {response.reason}".encode("utf-8"))
+                commlog.write(
+                    b"\n".join(
+                        to_wire(f"{x}: {response.headers[x]}") for x in response.headers
+                    )
+                )
+                commlog.write(b"\n\n")
+                ct = response.headers.get("Content-Type", "")
+                if response.tree is not None:
+                    commlog.write(
+                        to_wire(etree.tostring(response.tree, pretty_print=True))
+                    )
+                else:
+                    commlog.write(to_wire(response._raw))
+                commlog.write(b"\n")
 
         return response
+
+
+def auto_calendars(
+    configfile: str = f"{os.environ.get('HOME')}/.config/calendar.conf",
+    testconfig=False,
+    environment: bool = True,
+    config_data: dict = None,
+    config_name: str = None,
+) -> Iterable["Calendar"]:
+    """
+    This will replace plann.lib.findcalendars()
+    """
+    raise NotImplementedError("auto_calendars not implemented yet")
+
+
+def auto_calendar(*largs, **kwargs) -> Iterable["Calendar"]:
+    """
+    Alternative to auto_calendars - in most use cases, one calendar suffices
+    """
+    return next(auto_calendars(*largs, **kwargs), None)
+
+
+def auto_conn(
+    configfile: str = f"{os.environ.get('HOME')}/.config/calendar.conf",
+    testconfig=False,
+    environment: bool = True,
+    config_data: dict = None,
+    name: str = None,
+) -> "DAVClient":
+    """
+    Normally you would like to look into auto_calendars or
+    auto_calendar instead.  However, in some cases it's needed
+    with a DAVClient object rather than a Calendar object.
+
+    This function will yield a DAVClient object.  It will not try to
+    connect (see auto_calendars for that).  It will read configuration
+    from various sources, dependent on the parameters given, in this
+    order:
+
+    * Data from the given dict
+    * Environment variables prepended with "CALDAV_"
+    * Data from `./tests/conf.py` or `./conf.py` (this includes the possibility to spin up a test server)
+    * Configuration file.  Documented in the plann project as for now.  (TODO - move it)
+
+    """
+    if config_data:
+        return DAVClient(**config_data)
+
+    if testconfig:
+        sys.path.insert(0, "tests")
+        sys.path.insert(1, ".")
+        ## TODO: move the code from client into here
+        try:
+            from conf import client
+
+            try:
+                idx = int(name)
+                name = None
+            except ValueError:
+                idx = None
+            try:
+                conn = client(idx, name, **config_data)
+                if conn:
+                    return conn
+            except:
+                error.weirdness("traceback from client()")
+        except ImportError:
+            pass
+        finally:
+            sys.path = sys.path[2:]
+
+    if environment:
+        raise NotImplementedError(
+            "Not possible to configure the caldav server through environmental variables yet"
+        )
+
+    if configfile:
+        raise NotImplementedError(
+            "Support for configuration file not made yet (TODO: copy the code from the plann tool)"
+        )
