@@ -13,11 +13,21 @@ from datetime import timedelta
 from unittest import mock
 from urllib.parse import urlparse
 
-import caldav
 import icalendar
 import lxml.etree
 import pytest
 import vobject
+
+import caldav
+from caldav import Calendar
+from caldav import CalendarObjectResource
+from caldav import CalendarSet
+from caldav import DAVObject
+from caldav import Event
+from caldav import FreeBusy
+from caldav import Journal
+from caldav import Principal
+from caldav import Todo
 from caldav.davclient import DAVClient
 from caldav.davclient import DAVResponse
 from caldav.elements import cdav
@@ -28,15 +38,13 @@ from caldav.lib import url
 from caldav.lib.python_utilities import to_normal_str
 from caldav.lib.python_utilities import to_wire
 from caldav.lib.url import URL
-from caldav.objects import Calendar
-from caldav.objects import CalendarObjectResource
-from caldav.objects import CalendarSet
-from caldav.objects import DAVObject
-from caldav.objects import Event
-from caldav.objects import FreeBusy
-from caldav.objects import Journal
-from caldav.objects import Principal
-from caldav.objects import Todo
+
+## Note on the imports - those two lines are equivalent:
+# from caldav.objects import foo
+# from caldav import foo
+## This is due to a line like this in __init__.py:
+# from .objects import *
+## Said line should be deprecated at some point
 
 ## Some example icalendar data partly copied from test_caldav.py
 ev1 = """BEGIN:VCALENDAR
@@ -146,8 +154,117 @@ PRIORITY:1
 END:VTODO
 END:VCALENDAR"""
 
+## from https://github.com/python-caldav/caldav/issues/495
+recurring_task_response = """<d:multistatus xmlns:d="DAV:" xmlns:s="http://sabredav.org/ns" xmlns:cal="urn:ietf:params:xml:ns:caldav" xmlns:cs="http://calendarserver.org/ns/" xmlns:oc="http://owncloud.org/ns" xmlns:nc="http://nextcloud.org/ns">
+  <d:response>
+    <d:href>/remote.php/dav/calendars/oxi/personal/A9FFE819-5DDB-4947-A09C-308EEE5DA1F9.ics</d:href>
+    <d:propstat>
+      <d:prop>
+        <cal:calendar-data>BEGIN:VCALENDAR
+VERSION:2.0
+PRODID:+//IDN bitfire.at//ical4android (at.techbee.jtx)
+BEGIN:VTODO
+DTSTAMP:20250522T075151Z
+UID:8a8736b4-bc35-4085-a98b-89c2f52a5c51
+SEQUENCE:23
+CREATED:20250411T233004Z
+LAST-MODIFIED:20250522T075137Z
+SUMMARY:Clean Rosie filter
+DTSTART;VALUE=DATE:20250415
+RRULE:FREQ=WEEKLY;BYDAY=TU,FR
+PRIORITY:0
+END:VTODO
+BEGIN:VTODO
+DTSTAMP:20250522T075151Z
+UID:8a8736b4-bc35-4085-a98b-89c2f52a5c51
+SEQUENCE:1
+CREATED:20250411T234336Z
+LAST-MODIFIED:20250414T082134Z
+SUMMARY:Clean Rosie filter
+STATUS:COMPLETED
+DTSTART;VALUE=DATE:20250415
+RECURRENCE-ID;VALUE=DATE:20250415
+COMPLETED:20250414T082134Z
+PERCENT-COMPLETE:100
+PRIORITY:0
+END:VTODO
+BEGIN:VTODO
+DTSTAMP:20250522T075151Z
+UID:8a8736b4-bc35-4085-a98b-89c2f52a5c51
+SEQUENCE:1
+CREATED:20250411T234336Z
+LAST-MODIFIED:20250414T082134Z
+SUMMARY:Clean Rosie filter
+STATUS:CANCELLED
+DTSTART;VALUE=DATE:20250418
+RECURRENCE-ID;VALUE=DATE:20250418
+PRIORITY:0
+END:VTODO
+BEGIN:VTODO
+DTSTAMP:20250522T075151Z
+UID:8a8736b4-bc35-4085-a98b-89c2f52a5c51
+SEQUENCE:1
+CREATED:20250411T234336Z
+LAST-MODIFIED:20250414T082134Z
+SUMMARY:Clean Rosie filter
+STATUS:CANCELLED
+DTSTART;VALUE=DATE:20250422
+RECURRENCE-ID;VALUE=DATE:20250422
+PRIORITY:0
+END:VTODO
+BEGIN:VTODO
+DTSTAMP:20250522T075151Z
+UID:8a8736b4-bc35-4085-a98b-89c2f52a5c51
+SEQUENCE:1
+CREATED:20250411T234336Z
+LAST-MODIFIED:20250425T124511Z
+SUMMARY:Clean Rosie filter
+STATUS:COMPLETED
+DTSTART;VALUE=DATE:20250425
+RECURRENCE-ID;VALUE=DATE:20250425
+COMPLETED:20250425T124511Z
+PERCENT-COMPLETE:100
+PRIORITY:0
+END:VTODO
+BEGIN:VTODO
+DTSTAMP:20250522T075151Z
+UID:8a8736b4-bc35-4085-a98b-89c2f52a5c51
+SEQUENCE:1
+CREATED:20250411T234336Z
+LAST-MODIFIED:20250425T124511Z
+SUMMARY:Clean Rosie filter
+STATUS:CANCELLED
+DTSTART;VALUE=DATE:20250429
+RECURRENCE-ID;VALUE=DATE:20250429
+COMPLETED:20250425T124511Z
+PERCENT-COMPLETE:100
+PRIORITY:0
+END:VTODO
+BEGIN:VTODO
+DTSTAMP:20250522T075151Z
+UID:8a8736b4-bc35-4085-a98b-89c2f52a5c51
+SEQUENCE:1
+CREATED:20250411T234336Z
+LAST-MODIFIED:20250502T113705Z
+SUMMARY:Clean Rosie filter
+STATUS:COMPLETED
+DTSTART;VALUE=DATE:20250502
+RECURRENCE-ID;VALUE=DATE:20250502
+COMPLETED:20250502T113705Z
+PERCENT-COMPLETE:100
+PRIORITY:0
+END:VTODO
+END:VCALENDAR
+</cal:calendar-data>
+      </d:prop>
+      <d:status>HTTP/1.1 200 OK</d:status>
+    </d:propstat>
+  </d:response>
+</d:multistatus>
+"""
 
-def MockedDAVResponse(text):
+
+def MockedDAVResponse(text, davclient=None):
     """
     For unit testing - a mocked DAVResponse with some specific content
     """
@@ -156,17 +273,23 @@ def MockedDAVResponse(text):
     resp.reason = "multistatus"
     resp.headers = {}
     resp.content = text
-    return DAVResponse(resp)
+    return DAVResponse(resp, davclient)
 
 
-def MockedDAVClient(xml_returned):
+class MockedDAVClient(DAVClient):
     """
     For unit testing - a mocked DAVClient returning some specific content every time
     a request is performed
     """
-    client = DAVClient(url="https://somwhere.in.the.universe.example/some/caldav/root")
-    client.request = mock.MagicMock(return_value=MockedDAVResponse(xml_returned))
-    return client
+
+    def __init__(self, xml_returned):
+        self.xml_returned = xml_returned
+        DAVClient.__init__(
+            self, url="https://somwhere.in.the.universe.example/some/caldav/root"
+        )
+
+    def request(self, *largs, **kwargs):
+        return MockedDAVResponse(self.xml_returned)
 
 
 class TestExpandRRule:
@@ -264,6 +387,74 @@ class TestCalDAV:
         assert response.status == 200
         assert response.tree is None
 
+    def testSearchForRecurringTask(self):
+        client = MockedDAVClient(recurring_task_response)
+        calendar = Calendar(client, url="/calendar/issue491/")
+        mytasks = calendar.search(todo=True, expand=False)
+        assert len(mytasks) == 1
+        mytasks = calendar.search(
+            todo=True,
+            expand="client",
+            start=datetime(2025, 5, 5),
+            end=datetime(2025, 6, 5),
+        )
+        assert len(mytasks) == 9
+
+        ## It should not include the COMPLETED recurrences
+        mytasks = calendar.search(
+            todo=True,
+            expand="client",
+            start=datetime(2025, 1, 1),
+            end=datetime(2025, 6, 5),
+        )
+        assert len(mytasks) == 9
+
+    def testLoadByMultiGet404(self):
+        xml = """
+<D:multistatus xmlns:D="DAV:">
+  <D:response>
+    <D:href>/calendars/pythoncaldav-test/20010712T182145Z-123401%40example.com.ics</D:href>
+    <D:status>HTTP/1.1 404 Not Found</D:status>
+  </D:response>
+</D:multistatus>"""
+        client = MockedDAVClient(xml)
+        calendar = Calendar(client, url="/calendar/issue491/")
+        object = Event(url="/calendar/issue491/notfound.ics", parent=calendar)
+        with pytest.raises(error.NotFoundError):
+            object.load_by_multiget()
+
+    @mock.patch("caldav.davclient.requests.Session.request")
+    def testRequestCustomHeaders(self, mocked):
+        """
+        ref https://github.com/python-caldav/caldav/issues/285
+        also ref https://github.com/python-caldav/caldav/issues/385
+        """
+        mocked().status_code = 200
+        mocked().headers = {}
+        cal_url = "http://me:hunter2@calendar.møøh.example:80/"
+        client = DAVClient(
+            url=cal_url,
+            headers={"X-NC-CalDAV-Webcal-Caching": "On", "User-Agent": "MyCaldavApp"},
+        )
+        assert client.headers["Content-Type"] == "text/xml"
+        assert client.headers["X-NC-CalDAV-Webcal-Caching"] == "On"
+        ## User-Agent would be overwritten by some boring default in earlier versions
+        assert client.headers["User-Agent"] == "MyCaldavApp"
+
+    @mock.patch("caldav.davclient.requests.Session.request")
+    def testRequestUserAgent(self, mocked):
+        """
+        ref https://github.com/python-caldav/caldav/issues/391
+        """
+        mocked().status_code = 200
+        mocked().headers = {}
+        cal_url = "http://me:hunter2@calendar.møøh.example:80/"
+        client = DAVClient(
+            url=cal_url,
+        )
+        assert client.headers["Content-Type"] == "text/xml"
+        assert client.headers["User-Agent"].startswith("python-caldav/")
+
     @mock.patch("caldav.davclient.requests.Session.request")
     def testEmptyXMLNoContentLength(self, mocked):
         """
@@ -338,12 +529,14 @@ class TestCalDAV:
             "http://cal.example.com/home/bernard/calendars/"
         )
 
-    @mock.patch("caldav.CalendarObjectResource.is_loaded")
-    def testDateSearch(self, mocked):
+    def _load(self, only_if_unloaded=True):
+        self.data = todo6
+
+    @mock.patch("caldav.objects.CalendarObjectResource.load", new=_load)
+    def testDateSearch(self):
         """
         ## ref https://github.com/python-caldav/caldav/issues/133
         """
-        mocked.__bool__ = lambda self: True
         xml = """<xml><multistatus xmlns="DAV:">
 <response>
     <href>/principals/calendar/home@petroski.example.com/963/43B060B3-A023-48ED-B9E7-6FFD38D5073E.ics</href>
@@ -870,6 +1063,91 @@ END:VCALENDAR
             },
         }
 
+    def testHugeTreeParam(self):
+        """
+        With dealing with a huge XML response, such as event containing attachments, XMLParser will throw an exception
+        huge_tree parameters allows to handle this kind of events.
+        """
+
+        xml = """
+<multistatus xmlns="DAV:">
+  <response>
+    <href>/17149682/calendars/testcalendar-84439d0b-ce46-4416-b978-7b4009122c64/</href>
+    <propstat>
+      <prop>
+                </prop>
+      <status>HTTP/1.1 200 OK</status>
+    </propstat>
+    <propstat>
+      <prop>
+        <calendar-data xmlns="urn:ietf:params:xml:ns:caldav"/>
+      </prop>
+      <status>HTTP/1.1 404 Not Found</status>
+    </propstat>
+  </response>
+  <response>
+    <href>/17149682/calendars/testcalendar-84439d0b-ce46-4416-b978-7b4009122c64/20010712T182145Z-123401%40example.com.ics</href>
+    <propstat>
+      <prop>
+        <calendar-data xmlns="urn:ietf:params:xml:ns:caldav">
+
+BEGIN:VCALENDAR
+PRODID:-//MDaemon Technologies Ltd//MDaemon 21.5.2
+VERSION:2.0
+METHOD:PUBLISH
+BEGIN:VEVENT
+UID:
+ 040000008200E0007000B7101A82E0080000000050BF99B19D31
+SEQUENCE:0
+DTSTAMP:20230213T142930Z
+SUMMARY:This a summary of a very bug event
+DESCRIPTION:Description of this very big event
+LOCATION:Somewhere
+ORGANIZER:MAILTO:noreply@test.com
+PRIORITY:5
+ATTACH;VALUE=BINARY;ENCODING=BASE64;FMTTYPE=image/jpeg;
+ X-FILENAME=image001.jpg;X-ORACLE-FILENAME=image001.jpg:
+"""
+        xml += (
+            "gIyIoLTkwKCo2KyIjM4444449QEBAJjBGS0U+Sjk/QD3/2wBDAQsLCw8NDx0QEB09KSMpPT09\n"
+            * 153490
+        )
+        xml += """
+ /Z
+DTSTART;TZID="Europe/Paris":20230310T140000
+DTEND;TZID="Europe/Paris":20230310T150000
+END:VEVENT
+END:VCALENDAR
+</calendar-data>
+      </prop>
+      <status>HTTP/1.1 200 OK</status>
+    </propstat>
+  </response>
+</multistatus>
+"""
+        davclient = MockedDAVClient(xml)
+        resp = mock.MagicMock()
+        resp.headers = {"Content-Type": "text/xml"}
+        resp.content = xml
+
+        ## It seems like the huge_tree flag is not necessary in all
+        ## environments as of 2023-07.  Perhaps versioning issues with
+        ## the lxml library.
+        # davclient.huge_tree = False
+        # try:
+        #    import pdb; pdb.set_trace()
+        #    DAVResponse(resp, davclient=davclient)
+        #    assert False
+        # except Exception as e:
+        #    assert type(e) == lxml.etree.XMLSyntaxError
+
+        davclient.huge_tree = True
+        try:
+            DAVResponse(resp, davclient=davclient)
+            assert True
+        except:
+            assert False
+
     def testFailedQuery(self):
         """
         ref https://github.com/python-caldav/caldav/issues/54
@@ -971,6 +1249,19 @@ END:VCALENDAR
             == "Submit Quebec Income Tax Return for 2006"
         )
 
+    def testComponentSet(self):
+        cal_url = "http://me:hunter2@calendar.example:80/"
+        client = DAVClient(url=cal_url)
+        target = Event(client, data=evr)
+
+        ## Creating some dummy data such that the target has more than one subcomponent
+        target.expand_rrule(start=datetime(1996, 10, 10), end=datetime(1999, 12, 12))
+        assert len(target.icalendar_instance.subcomponents) == 3
+
+        ## The following should not fail within _set_icalendar_component
+        target.icalendar_component = icalendar.Todo.from_ical(todo).subcomponents[0]
+        assert len(target.icalendar_instance.subcomponents) == 1
+
     def testTodoDuration(self):
         cal_url = "http://me:hunter2@calendar.example:80/"
         client = DAVClient(url=cal_url)
@@ -1152,7 +1443,7 @@ END:VCALENDAR
 
     def test_calendar_comp_class_by_data(self):
         calendar = Calendar()
-        for (ical, class_) in (
+        for ical, class_ in (
             (ev1, Event),
             (todo, Todo),
             (journal, Journal),
@@ -1182,12 +1473,12 @@ END:VCALENDAR
         """
         cal_url = "http://me:hunter2@calendar.example:80/"
         with DAVClient(url=cal_url) as client:
-            assert client.extract_auth_types("Basic\n") == ["basic"]
-            assert client.extract_auth_types("Basic") == ["basic"]
-            assert client.extract_auth_types('Basic Realm=foo;charset="UTF-8"') == [
+            assert client.extract_auth_types("Basic\n") == {"basic"}
+            assert client.extract_auth_types("Basic") == {"basic"}
+            assert client.extract_auth_types('Basic Realm=foo;charset="UTF-8"') == {
                 "basic"
-            ]
-            assert client.extract_auth_types("Basic,dIGEST Realm=foo") == [
+            }
+            assert client.extract_auth_types("Basic,dIGEST Realm=foo") == {
                 "basic",
                 "digest",
-            ]
+            }
