@@ -3,21 +3,20 @@ import re
 import uuid
 from datetime import datetime
 from datetime import timedelta
+from datetime import timezone
 from unittest import TestCase
 
 import icalendar
 import pytest
-import pytz
 import vobject
+
 from caldav.lib import vcal
 from caldav.lib.python_utilities import to_normal_str
 from caldav.lib.python_utilities import to_wire
 from caldav.lib.vcal import create_ical
 from caldav.lib.vcal import fix
 
-# from datetime import timezone
-# utc = timezone.utc
-utc = pytz.utc
+utc = timezone.utc
 
 # example from http://www.rfc-editor.org/rfc/rfc5545.txt
 ev = """BEGIN:VCALENDAR
@@ -135,6 +134,32 @@ class TestVcal(TestCase):
         There is an obscure function lib.vcal that attempts to fix up
         known ical standard breaches from various calendar servers.
         """
+        non_broken_ical = [
+            """BEGIN:VCALENDAR
+VERSION:2.0
+PRODID:-//python-caldav//caldav//en_DK
+BEGIN:VEVENT
+SUMMARY:Foo bar blah
+ blub
+DTSTART:20230313T084500Z
+DURATION:PT30M
+DTSTAMP:20230312T215907Z
+UID:1c9bba3e-c121-11ed-bf96-982cbcdd642c
+CATEGORIES:oslo
+END:VEVENT
+END:VCALENDAR
+""",
+            ## Next one contains a DTSTAMP before BEGIN:VEVENT
+            ## Doesn’t make sense, but valid, and more importantly,
+            ## not failing during the `fix` call.
+            """DTSTAMP:20210205T101751Z
+BEGIN:VEVENT
+UID:20200516T060000Z-123401@example.com
+SUMMARY:Do the needful
+DTSTART:20210517T060000Z
+END:VEVENT
+""",
+        ]
         broken_ical = [
             ## This first one contains duplicated DTSTAMP in the event data
             """BEGIN:VCALENDAR
@@ -224,12 +249,42 @@ TRIGGER;RELATED=START:-PT15M
 DESCRIPTION:Reminder
 END:VALARM
 END:VEVENT
+END:VCALENDAR""",  ## next has a completed date.  It should be a timestamp according to the RFC
+            """BEGIN:VCALENDAR
+VERSION:2.0
+PRODID:-//Example Corp.//CalDAV Client//EN
+BEGIN:VTODO
+UID:20070313T123432Z-456553@example.com
+DTSTAMP:20070313T123432Z
+COMPLETED;VALUE=DATE:20070501
+SUMMARY:Submit Quebec Income Tax Return for 2006
+CLASS:CONFIDENTIAL
+CATEGORIES:FAMILY,FINANCE
+STATUS:NEEDS-ACTION
+END:VTODO
+END:VCALENDAR""",  ## Again, but implicit instead of explicit date
+            """BEGIN:VCALENDAR
+VERSION:2.0
+PRODID:-//Example Corp.//CalDAV Client//EN
+BEGIN:VTODO
+UID:20070313T123432Z-456553@example.com
+DTSTAMP:20070313T123432Z
+COMPLETED:20070501
+SUMMARY:Submit Quebec Income Tax Return for 2006
+CLASS:CONFIDENTIAL
+CATEGORIES:FAMILY,FINANCE
+STATUS:NEEDS-ACTION
+END:VTODO
 END:VCALENDAR""",
         ]  ## todo: add more broken ical here
 
         for ical in broken_ical:
             ## This should raise error
-            with pytest.raises(vobject.base.ValidateError):
+            # with pytest.raises(vobject.base.ValidateError):
+            with pytest.raises(Exception):
                 vobject.readOne(ical).serialize()
             ## This should not raise error
             vobject.readOne(vcal.fix(ical)).serialize()
+
+        for ical in non_broken_ical:
+            assert vcal.fix(ical) == ical
